/**  File:        CADstuff.cpp
 **  Package:     none
 **  Created by:  markan
 **  Modified by: markan, wml
 **  Description: C++ datastructure for UG files
 **
 **  See CADstuff.h for more extensive information.
 **/

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <sstream>
#include <list>
#include <vector>
#include <algorithm>
#include <utility>
#include <map>
#include "CADstuff.h"

using namespace std;

// this is really, truly awful -- but some compilers would do it for us
//   silently, so better to do it ourselves, for now
#include "support.cc"

#ifndef MAXFLOAT
#define MAXFLOAT ((float)3.40282346638528860e+38)
#endif


unsigned int alg_table[ALG_ARRAY_SIZE_1][ALG_ARRAY_SIZE_2];


bool verbose = true;

/****************************************************************************/
/**                           METHOD DEFINITIONS                           **/
/****************************************************************************/

// ** "ug_taggable" method definitions

int ug_taggable::check(int name) {
  /*
  if (names.contains(name))
    return values.at(names.find(name));
  return 0;
  */
  list< pair<int,int> >::iterator it;
  
  it = find_if(names_values.begin(), names_values.end(), first_equals(name));
  //it = find(names.begin(), names.end(), name);
  //if (*it == name) return *it;
  //else return 0;  
  if (it != names_values.end()) return it->second;
  else return 0;
}

void ug_taggable::tag(int name, int value) {
  /*
  if (names.contains(name)) {
    values.set(names.find(name), value);
  } else {
    names.add(name);
    values.add(value);
  }
  */
  list< pair<int,int> >::iterator it;

  it = find_if(names_values.begin(), names_values.end(), first_equals(name));
  if (it != names_values.end()) {
    it->second = value;
  } else {
    names_values.push_back( pair<int,int>(name, value) );
  }
}

// ** ug_vert definitions

ug_vert::ug_vert(ug_CADSpace *where,
                 double xc, double yc, double zc, ug_def *d)
{
  stringstream s;

  home = where;
  //sprintf(name, "n%d", home->unusedVertNum());

  s << 'n' << home->unusedVertNum();
  name = s.str();

  myRegion = d;
  myRegion->add(this);
  x = xc;
  y = yc;
  z = zc;
}

ug_vert::ug_vert(ug_CADSpace *where, char *n,
                 double xc, double yc, double zc, ug_def *d)
{

  home = where;
  //strcpy(name, n);

  name = string(n);

  myRegion = d;
  myRegion->add(this);
  x = xc;
  y = yc;
  z = zc;
}

/**  Method:      *ug_vert::get_name( void )
 **  Description: Return the name array.
 **/
char *ug_vert::get_name() { return (char*) name.c_str(); }

void ug_vert::add(ug_edge *e) {
  //aEdges.add(e);
  aEdges.push_back(e);
}

void ug_vert::add(ug_wire *w) {
  //pWires.add(w);
  pWires.push_back(w);
}

ug_vert *ug_vert::dup(double xc, double yc, double zc, ug_def *d) {
  return home->addNewVert(x + xc, y + yc, z + zc, d);
}

bool ug_vert::match(ug_vert *v) {
  if (x == v->x && y == v->y && z == v->z) return true;
  return false;
}

ug_vert *ug_vert::dupUsingAxis(ug_edge *axis, double xc, double yc, ug_def *d) {
  double normX, normY;

  if (axis->length() == 0) {
    printf("uhoh"); //delete eventually
    return NULL;
  }
  normX = (axis->aVerts[1]->x - axis->aVerts[0]->x) / axis->length();
  normY = (axis->aVerts[1]->y - axis->aVerts[0]->y) / axis->length();

  return home->addNewVert(x + normX * xc - normY * yc, y + xc * normY + yc * normX, z, d);
}

double ug_vert::distance(ug_vert *other) {
  return sqrt((x - other->x) * (x - other->x) + (y - other->y) * (y - other->y));
}

void ug_vert::set(double xc, double yc, double zc) {
  x = xc;
  y = yc;
  z = zc;
}

ug_vert::~ug_vert() {
}

/* void ug_vert::setRegion (ug_region *region) {
  //  if (myRegion != NULL) myRegion->remove(this); add back!
  if (myRegion == region || (myRegion == home->nullRegion && region == NULL)) return;
  if (region == NULL)
    myRegion = home->nullRegion;
  else
    myRegion = region;
  myRegion->add(this);
  } */


// ** ug_edge definitions

ug_edge::ug_edge(ug_CADSpace *where, ug_vert *v1c, ug_vert *v2c, ug_wire *parentc) {
  int i;

  aVerts[0] = v1c;
  aVerts[1] = v2c;
  parent = parentc;
  home = where;

  for (i = 0; i < 2; i++) aVerts[i]->add(this);
}

ug_edge *ug_edge::dup() {
  return new ug_edge(home, aVerts[0], aVerts[1], parent);
}

bool ug_edge::contains(ug_vert *v) {
  //  printf("td\n");
  //  if (aVerts[0] == NULL || aVerts[1] == NULL) printf("a\n");
  //  printf("con: %f %f %f\n", aVerts[0]->distance(v), aVerts[1]->distance(v), aVerts[0]->distance(aVerts[1]));
  if (aVerts[0]->distance(v) + aVerts[1]->distance(v) == aVerts[0]->distance(aVerts[1])) return true;
  return false;
}

bool ug_edge::isAbove(ug_vert *v) {
  double x1, x2, x3, y1, y2, y3;

  if (v == NULL) return false;

  x1 = aVerts[0]->x; y1 = aVerts[0]->y;
  x2 = aVerts[1]->x; y2 = aVerts[1]->y;
  x3 = v->x; y3 = v->y;

  if (x1 * y2 + x2 * y3 + x3 * y1 - x2 * y1 - x3 * y2 - x1 * y3 < 0) return true;
  return false;
}

bool ug_edge::isParallel(ug_edge *other) {
  double x1, x2, y1, y2;

  x1 = aVerts[1]->x - aVerts[0]->x;
  y1 = aVerts[1]->y - aVerts[0]->y;
  x2 = other->aVerts[1]->x - other->aVerts[0]->x;
  y2 = other->aVerts[1]->y - other->aVerts[0]->y;

  if (fabs(x1 * y2 - x2 * y1) < TOL) return true;
  return false;
}

bool ug_edge::matchDir(ug_edge *other) {
  if (!isParallel(other)) return false;
  if (project(other->aVerts[1]) > project(other->aVerts[0])) return true;
  return false;
}

// ug_edge::overlaps - determine if two edges (assumed to be parallel and collinear) overlap

bool ug_edge::overlaps(ug_edge *other) {
  double u1, u2, u3, u4, mn, mx;

  u1 = project(other->aVerts[0]);
  u2 = project(other->aVerts[1]);
  u3 = other->project(aVerts[0]);
  u4 = other->project(aVerts[1]);

  mn = u1;
  if (u2 < mn) mn = u2;
  if (0 < mn) mn = 0;
  if (length() < mn) mn = length();
  mx = u1;
  if (u2 > mx) mx = u2;
  if (0 > mx) mx = 0;
  if (length() > mx) mx = length();

  //  printf("[overlaps] %f %f %f\n", mx - mn, length(), other->length());

  if ((mx - mn) >= length() + other->length() - .01) // fudging
    return false;
  return true;
}

double ug_edge::distance(ug_edge *other) {
  double a, b;

  a = aVerts[0]->distance(other->aVerts[0]);
  b = project(other->aVerts[0]);

  if (a * a - b * b < 0) {
    if (verbose) printf("Warning: returned 0 for sqrt(%f)\n", a * a - b * b);
    return 0;
  }

  return sqrt(a * a - b * b);
}

double ug_edge::distance(ug_vert *other) {
  double a, b;

  a = aVerts[0]->distance(other);
  b = project(other);

  return sqrt(a * a - b * b);
}

double ug_edge::length() {
  double x, y;
  x = aVerts[1]->x - aVerts[0]->x;
  y = aVerts[1]->y - aVerts[0]->y;
  return sqrt(x * x + y * y);
}

ug_edge *ug_edge::overlap(ug_edge *other) {
  double u1, u2;
  ug_vert *v1, *v2;

  u1 = project(other->aVerts[0]);
  u2 = project(other->aVerts[1]);
  
  if (u2 > u1) {
    if (u1 > 0)
      v1 = other->aVerts[0];
    else
      v1 = aVerts[0];
    if (u2 > length())
      v2 = aVerts[1];
    else
      v2 = other->aVerts[1];
  } else {
    if (u2 > 0)
      v1 = other->aVerts[1];
    else
      v1 = aVerts[0];
    if (u1 > length())
      v2 = aVerts[1];
    else
      v2 = other->aVerts[0];
  }
      
  return new ug_edge(home, v1, v2, NULL);
}

bool ug_edge::coincides(ug_edge *other) {
  double max, u1, u2, l1, l2, min = 0;

  max = l1 = length();
  l2 = other->length();

  if (l1 == 0 || l2 == 0) return false;

  if (isParallel(other) && distance(other) < TOL) {
    u1 = project(other->aVerts[0]);
    u2 = project(other->aVerts[1]);
    //    printf("tf %f %f %f\n", u1, u2, length());
    
    if (u1 < min) min = u1; if (u2 < min) min = u2;
    if (u1 > max) max = u1; if (u2 > max) max = u2;

    if (max - min < l1 + l2) {return true;}
  }
  return false;
}

double ug_edge::project(ug_vert *v) {
  double x1, y1, x2, y2, l;

  l = length();
  if (l == 0) return 0;

  x1 = aVerts[1]->x - aVerts[0]->x;
  y1 = aVerts[1]->y - aVerts[0]->y;
  x2 = v->x - aVerts[0]->x;
  y2 = v->y - aVerts[0]->y;

  return (x1 * x2 + y1 * y2) / l;
}

ug_wire *ug_edge::extrude(double s, double h, ug_def *d) {
  ug_wire *ret;
  ug_layer *ltu;

  if (parent) {
    ltu = parent->myLayer;
  } else {
    ltu = home->getLayerByNumber(0);
  }

  ret = home->addNewWire(ltu, d);
  ret->pushVert(aVerts[1]->dup(0, 0, s, d));
  ret->pushVert(aVerts[0]->dup(0, 0, s, d));
  ret->pushVert(aVerts[0]->dup(0, 0, h, d));
  ret->pushVert(aVerts[1]->dup(0, 0, h, d));
  ret->close();

  return ret;
}

/* vertex *line::left(vertex *v1t, vertex *v2t) {
  double x0, y0, x1, y1, x2, y2;

  x0 = v2->x - v1->x;
  y0 = v2->y - v1->y;
  x1 = v1t->x - v1->x;
  y1 = v1t->y - v1->y;
  x2 = v2t->x - v1->x;
  y2 = v2t->y - v1->y;

  if (x0 * x1 + y0 * y1 < x0 * x2 + y0 * y2)
    return v1t;
  return v2t;
}

vertex *line::right(vertex *v1t, vertex *v2t) {
  double x0, y0, x1, y1, x2, y2;

  x0 = v2->x - v1->x;
  y0 = v2->y - v1->y;
  x1 = v1t->x - v1->x;
  y1 = v1t->y - v1->y;
  x2 = v2t->x - v1->x;
  y2 = v2t->y - v1->y;

  if (x0 * x1 + y0 * y1 > x0 * x2 + y0 * y2)
    return v1t;
  return v2t;
} */

/*bool ug_edge::intersects(double *x = NULL, double *y = NULL) {
  if (ug_edge->aVerts[0]->x == ug_edge->a
  }*/

void ug_edge::flip() {
     ug_vert *t;

     t = aVerts[0];
     aVerts[0] = aVerts[1];
     aVerts[1] = t;
}

// ** ug_wire definitions

ug_wire::ug_wire(char *n) {
  isFace = false;
  myLayer = NULL;
  myRegion = NULL;
  home = NULL;
  strcpy(name, n);
}

ug_wire::ug_wire(ug_CADSpace *where, ug_layer *l, ug_def *d) {
  isFace = false;
  myLayer = l;
  myRegion = d;
  myRegion->add(this);
  home = where;
  sprintf(name, "wn%d", home->unusedWireNum());
}

ug_wire *ug_wire::unfill() {
  ug_wire *ret;

  assert(isFace);

  ret = dup(0, 0, 0, myRegion);
  ret->isFace = false;
  ret->pushVert(ret->getVert(0));

  return ret;
}

ug_wire *ug_wire::dupClean(ug_def *d) {
  int i, j;
  ug_wire *ret;
  ret = home->addNewWire(myLayer, d);

  i = 0;
  if (isFace) {
    assert(numEdges() == numVerts());

    // Find a safe place to start

    for (i = 0; i < numEdges(); i++)
      if (!getEdge(i)->isParallel(getEdge((i - 1 + numEdges()) % numEdges()))) break;

    assert(i != numEdges());

    j = i;

    do {
      if (!getEdge(i)->isParallel(getEdge((i - 1 + numEdges()) % numEdges()))) ret->pushVert(getVert(i)->dup(0, 0, 0, d));
      i = (i + 1) % numEdges();
    } while (i != j);

    ret->close();
  } else {
    ret = dup(0, 0, 0, d);
  }

  return ret;
}

ug_wire *ug_wire::dup(double xc, double yc, double zc, ug_def *d) {
  ug_wire *ret = home->addNewWire(myLayer, d);
  for(int i = 0; i < numVerts(); i++)
    ret->pushVert(getVert(i)->dup(xc, yc, zc, d));
  if (isFace) ret->close();
  return ret;
}

ug_vert *ug_wire::firstJoin(ug_wire *otherWire, bool *backwards) {
  int i, j;
  ug_edge *cEdge, *otherEdge;
  double u1, u2;

  (*backwards) = false;

  //  printf("%d %d\n", numEdges(), otherWire->numEdges());

  for (i = 0; i < numEdges(); i++) {
    cEdge = getEdge(i);
    if (cEdge->length() == 0) continue;

    for (j = 0; j < otherWire->numEdges(); j++) {
      otherEdge = otherWire->getEdge(j);
      if (otherEdge->length() == 0) continue;
      
      if (cEdge->coincides(otherEdge)) {
	u1 = cEdge->project(otherEdge->aVerts[0]);
	u2 = cEdge->project(otherEdge->aVerts[1]);
	assert (u1 != u2);

	if (u2 < u1) {
	  (*backwards) = true;
	  if (u2 <= 0) {
	    return cEdge->aVerts[0];
	  } else {
	    return otherEdge->aVerts[1];
	  }
	} else {
	  if (u1 <= 0) {
	    return cEdge->aVerts[0];
	  } else {
	    return otherEdge->aVerts[0];
	  }
	}
      }
    }
  }
  return NULL;
}

ug_vert *ug_wire::firstSplit(ug_wire *fromThis, bool backwards) {
  int i = 0;
  int j = 0;
  ug_edge *heading;
  ug_wire *flipped;

  flipped = fromThis->dup(0, 0, 0, myRegion);
  if (backwards) flipped->flip();

  for (j = 0; j < fromThis->numEdges(); j++) {
    if (flipped->getEdge(j)->contains(getVert(0)) && !flipped->getEdge(j)->aVerts[1]->match(getVert(0)) && flipped->getEdge(j)->matchDir(getEdge(0))) break;
  }

  printf("[firstSplit] Run began at edge %d\n", j);

  assert (j != flipped->numEdges());

  heading = getEdge(i);

  //  printf("%d\n", j);

  while (true) {
    //getEdge(i)->aVerts[1];
    //    printf("%d %d %f %f\n"
    //    printf("%f %f\n", heading->project(getEdge(i)->aVerts[1]), heading->project(flipped->getEdge(j)->aVerts[1]));
    //    printf("d\n");  
    if (heading->project(getEdge(i)->aVerts[1]) < heading->project(flipped->getEdge(j)->aVerts[1])) {
      //      printf("a\n");
      i++;
      if (i == numEdges()) return NULL;
      if (!heading->matchDir(getEdge(i))) return getEdge(i)->aVerts[0];
    } else if (heading->project(getEdge(i)->aVerts[1]) > heading->project(flipped->getEdge(j)->aVerts[1])) {
      //      printf("b\n");
      if (!flipped->isFace && j == flipped->numEdges() - 1) return flipped->getEdge(j)->aVerts[1];
      j = (j + 1) % flipped->numEdges();
      if (!heading->matchDir(flipped->getEdge(j))) return flipped->getEdge(j)->aVerts[0];
    } else {
      //      printf("c\n");
      i++;
      if (i == numEdges()) return NULL;
      heading = getEdge(i);
      if (heading->project(flipped->getEdge((j + 1) % flipped->numEdges())->aVerts[1]) < 0) return getEdge(i)->aVerts[0];
    }
  }
}

void ug_wire::split(ug_vert *v, ug_wire **newWire1, ug_wire **newWire2) {
  int i;
  ug_edge *cEdge = NULL;
  bool onFirst = true;

  (*newWire1) = home->addNewWire(myLayer, myRegion);
  (*newWire2) = home->addNewWire(myLayer, myRegion);
  
  for (i = 0; i < numEdges(); i++) {
    cEdge = getEdge(i);
    if (onFirst) (*newWire1)->pushVert(cEdge->aVerts[0]);
    if (!onFirst) (*newWire2)->pushVert(cEdge->aVerts[1]);
    if (cEdge->contains(v) && onFirst == true) {
      //      printf("c:%d\n", i);
      (*newWire1)->pushVert(v);
      (*newWire2)->pushVert(v);
      if (!v->match(cEdge->aVerts[1])) (*newWire2)->pushVert(cEdge->aVerts[1]);
      onFirst = false;
    }
  }

  (*newWire1)->printme();

  if (onFirst) {
    (*newWire1)->pushVert(cEdge->aVerts[1]);
    if (isFace) (*newWire1)->close();
  } else {
    if (isFace) (*newWire2)->pushVert(getVert(0));
    //    printf("split error: tried to split closed loop\n");
  }
}



void ug_wire::replaceVert(ug_vert *preV, ug_vert *newV)
{
  
}




double ug_wire::length() {
  double ret = 0;
  int i;

  for (i = 0; i < numEdges(); i++)
    ret += getEdge(i)->length();

  return ret;
}

double ug_wire::area() {
  double px, py, ret = 0;
  ug_vert *cVert;
  
  cVert = getVert(0);
  px = cVert->x;
  py = cVert->y;
  
  for(int i = 1; i <= numVerts(); i++) {
    cVert = getVert(i % numVerts());
    ret += (cVert->x - px) * (cVert->y + py) / 2;
    px = cVert->x;
    py = cVert->y;
  }
  return ret;
}

int ug_wire::numVerts(){
  return vertList.size;
}

int ug_wire::numEdges() {
  return vertList.size - 1 + isFace;
}

// this should probably be deprecated
ug_edge *ug_wire::getEdge(int i) {
  return edgeList.at(i);
}

// ug_wire::printme() - print the wire's properties to stdout, used for diagnostics

void ug_wire::printme() {
  printf("ug_wire %s has %d vertices:\n", name, numVerts());
  for (int i = 0; i < numVerts(); i++) {
    printf("%f %f %f\n", getVert(i)->x, getVert(i)->y, getVert(i)->z);
    //    if (i < numEdges()) printf("%f %f %f\n", getEdge(i)->aVerts[0]->x, getEdge(i)->aVerts[0]->y, getEdge(i)->aVerts[0]->z);
  }
  if (isFace) printf("isFace\n");
}

ug_sel *ug_wire::extrude(double s, double h, ug_def *d) {
  ug_sel *ret;

  ret = new ug_sel();
  for(int i = 0; i < numVerts() - 1 + isFace; i++)
    ret->add(getEdge(i)->extrude(s, h, d));

  return ret;
}

ug_vert *ug_wire::getVert(int i){
  return vertList.at(i);
}

void ug_wire::pushVert(ug_vert *vertc) {

  // this is not something we want the UG library to do for us, comment it out
  /*
  if (numVerts() > 0 && vertc->match(getVert(numVerts() - 1))) return;
  */

  vertList.add(vertc);
  /*
  if (numVerts() > 1) edgeList.add(new ug_edge(home, getVert(numVerts() - 2), vertc, this));
  */
  if (numVerts() > 1)
    edgeList.push_back(new ug_edge(home, getVert(numVerts()-2), vertc, this));
  vertc->add(this);
  assert(myRegion == vertc->myRegion || home == NULL);
}

void ug_wire::close() {
  //edgeList.add(new ug_edge(home, getVert(numVerts() - 1), getVert(0), this));
  edgeList.push_back(new ug_edge(home, getVert(numVerts()-1),
                                 getVert(0), this));
  isFace = true;
}

void ug_wire::open() {
  isFace = false;
  vertList.add(getVert(0));
}

// Seems like it shouldn't work...why?

void ug_wire::flip() {
  h_table<ug_vert> tempList;
  //list<ug_edge *> tempList2;
  vector <ug_edge*>tempList2;
  int i;

  for (i = numVerts() - 1; i >= 0; i--) {
    tempList.add(getVert(i));
  }

  for (i = numVerts() - 2 + isFace; i >= 0; i--) {
    //tempList2.add(getEdge(i));
    tempList2.push_back(getEdge(i));
    getEdge(i)->flip();
  }

  vertList = tempList;
  edgeList = tempList2;
}

void ug_wire::setLayer(ug_layer *layer) {
  myLayer = layer;
}

/* void ug_wire::setRegion(ug_region *region) {
  //  if (myRegion != NULL) myRegion->remove(this); add back!
  if (region == NULL)
    myRegion = home->nullRegion;
  else
    myRegion = region;
  myRegion->add(this);

  for (int i = 0; i < numVerts(); i++)
    getVert(i)->setRegion(region);
    } */

bool ug_wire::matchOrient(ug_wire *w) {
  ug_edge *cEdge, *e;

  for (int i = 0; i < numEdges(); i++) {
    cEdge = getEdge(i);
    for (int j = 0; j < w->numEdges(); j++) {
      e = w->getEdge(j);
      if (cEdge->isParallel(e) && cEdge->overlaps(e) && e->distance(cEdge) < TOL) {
	tag(ORIENTED, 1);
	if (cEdge->project(e->aVerts[1]) - cEdge->project(e->aVerts[0]) < 0) flip();
	return true;
      }
    }
  }

  return false;
}

double ug_wire::cmx() {
  double ret = 0;

  for (int i = 0; i < numVerts(); i++)
    ret += getVert(i)->x;

  //  printf("%f", ret/numVerts());
  return ret/numVerts();
}

double ug_wire::cmy() {
  double ret = 0;

  for (int i = 0; i < numVerts(); i++)
    ret += getVert(i)->y;

  return ret/numVerts();
}

/**  Method:      *ug_wire::get_name( void )
 **  Description: Return the string containing the name of the wire.
 **/
char *ug_wire::get_name(){ return name; }

// ** ug_sel method definitions

ug_sel::ug_sel() {
}

ug_sel::ug_sel(char *n) {
  strcpy(name, n);
}

ug_sel *ug_sel::dup(double dx, double dy, double dz) {
  ug_sel *ret = new ug_sel();

  for (int i = 0; i < numWires(); i++)
    ret->add(getWire(i)->dup(dx, dy, dz, getWire(i)->myRegion));

  return ret;
}

int ug_sel::numVerts() {
  return vertList.size;
}

int ug_sel::numEdges() {
  return edgeList.size();
}

int ug_sel::numWires() {
  return wireList.size;
}

ug_vert *ug_sel::getVert(int i) {
  return vertList.at(i);
}

ug_edge *ug_sel::getEdge(int i) {
  return edgeList.at(i);
}

ug_wire *ug_sel::getWire(int i) {
  return wireList.at(i);
}

void ug_sel::align(ug_sel *toMe, double *dx, double *dy) {
  ug_vert *cornerVert[2][4];
  double cornerVal[2][4];
  double testCornerVal;
  ug_vert *cVert;
  int i, j;
  double bestEval = 0;
  int bestIndex = 0;
  int e;

  if (numVerts() == 0 || toMe->numVerts() == 0) {
    *dx = 0;
    *dy = 0;
    return;
  }

  for (i = 0; i < numVerts(); i++) {
    cVert = getVert(i);

    for (j = 0; j < 4; j++) {
      testCornerVal = ((j & 2) - 1) * cVert->x + ((j & 1) * 2 - 1) * cVert->y;
      if (i == 0 || testCornerVal > cornerVal[0][j]) {
	cornerVal[0][j] = testCornerVal;
	cornerVert[0][j] = cVert;
      }
    }
  }

  for (i = 0; i < toMe->numVerts(); i++) {
    cVert = toMe->getVert(i);

    for (j = 0; j < 4; j++) {
      testCornerVal = ((j & 2) - 1) * cVert->x + ((j & 1) * 2 - 1) * cVert->y;
      if (i == 0 || testCornerVal > cornerVal[1][j]) {
	cornerVal[1][j] = testCornerVal;
	cornerVert[1][j] = cVert;
      }
    }
  }
  printf("gsf\n");

  for (i = 0; i < 4; i++) {
    printf("%f\n", cornerVert[1][i]->x);
    printf("%f\n", cornerVert[0][i]->x);
    printf("%f\n", cornerVert[1][i]->y);
    printf("%f\n", cornerVert[0][i]->y);
    e = evalAlign(toMe, cornerVert[1][i]->x - cornerVert[0][i]->x, cornerVert[1][i]->y - cornerVert[0][i]->y);
    if (e > bestEval) {
      bestEval = e;
      bestIndex = i;
    }
  }

  *dx = cornerVert[1][bestIndex]->x - cornerVert[0][bestIndex]->x;
  *dy = cornerVert[1][bestIndex]->y - cornerVert[0][bestIndex]->y;
}

int ug_sel::evalAlign(ug_sel *toMe, double dx, double dy) {
  int ret = 0;
  ug_vert *cVert;

  for (int i = 0; i < numVerts(); i++) {
    cVert = getVert(i);
    cVert->set(cVert->x + dx, cVert->y + dy, cVert->z);
    for (int j = 0; j < toMe->numVerts(); j++) {
      if (cVert->distance(toMe->getVert(j)) < NEARTOL) {
	ret++;
	break;
      }
    }
    cVert->set(cVert->x - dx, cVert->y - dy, cVert->z);
  }

  return ret;
}

void ug_def::addSelfToCADSpace(ug_CADSpace *ret) {
  ug_wire *cWire, *nWire;
  ug_vert *cVert, *nVert;
  ug_inline *cInline;
  ug_instance *cInstance, *nInstance;
  ug_def *nDef;
  int i;

  nDef = ret->addNewRegion(name);

  printf("Def %s adding itself\n", nDef->name);

  for (i = 0; i < numWires(); i++) {
    cWire = getWire(i);

    if (ret->getLayer(cWire->myLayer->name) == NULL)
      ret->addNewLayer(cWire->myLayer->name, cWire->myLayer->r, cWire->myLayer->g, cWire->myLayer->b, cWire->myLayer->c);

    nWire = ret->addNewWire(ret->getLayer(cWire->myLayer->name), nDef);

    for (int j = 0; j < cWire->numVerts(); j++) {
      cVert = cWire->getVert(j);
      nVert = ret->addNewVert(cVert->x, cVert->y, cVert->z, nDef);
      nWire->pushVert(nVert);
    }

    if (cWire->isFace) nWire->close();
  }

  for (i = 0; i < numInstances(); i++) {
    cInstance = getInstance(i);
    assert (cInstance->basedOn != NULL);

    if (ret->getRegion(cInstance->basedOn->name) == NULL)
      cInstance->basedOn->addSelfToCADSpace(ret);

    nInstance = ret->addNewInstance(nDef);
    assert(ret->getRegion(cInstance->basedOn->name) != NULL);
    nInstance->setObject(ret->getRegion(cInstance->basedOn->name));
    printf("%p\n", (void*)ret->getRegion(cInstance->basedOn->name));
    nInstance->setLayer(ret->getLayer(cInstance->myLayer->name));
    nInstance->makeTransform((char*)cInstance->getTransform());
    nInstance->setName((char*)cInstance->getName());
  }

  for (i = 0; i < numInlines(); i++) {
    cInline = getInline(i);
    ret->include(cInline->getFilename(), nDef);
  }
}

/* ug_CADSpace *ug_def::toCADSpace() {
  ug_CADSpace *ret = new ug_CADSpace();
  ug_wire *cWire;
  ug_wire *nWire;
  ug_vert *cVert;
  ug_vert *nVert;

  for (int i = 0; i < numWires(); i++) {
    cWire = getWire(i);

    if (ret->getLayer(cWire->myLayer->name) == NULL)
      ret->addNewLayer(cWire->myLayer->name, cWire->myLayer->r, cWire->myLayer->g, cWire->myLayer->b, cWire->myLayer->c);

    if (cWire->myRegion != NULL && ret->getRegion(cWire->myRegion->name) == NULL)
      ret->addNewRegion(cWire->myRegion->name);

    nWire = ret->addNewWire(ret->getLayer(cWire->myLayer->name), ret->getRegion(cWire->myRegion->name));

    for (int j = 0; j < cWire->numVerts(); j++) {
      cVert = cWire->getVert(j);
      nVert = ret->addNewVert(cVert->x, cVert->y, cVert->z, ret->getRegion(cVert->myRegion->name));
      nWire->pushVert(nVert);
    }

    if (cWire->isFace) nWire->close();
  }

  return ret;
  } */

ug_vertOfWire ug_sel::anyMatchingVert(ug_vert *v, ug_wire *except) {
  ug_wire *cWire;
  ug_vert *cVert;
  ug_vertOfWire ret;

  for (int i = 0; i < numWires(); i++) {
    cWire = getWire(i);
    for (int j = 0; j < cWire->numVerts(); j++) {
      cVert = cWire->getVert(j);
      if (cVert->match(v) && cWire != except) {
	ret.v = cVert;
	ret.w = cWire;
	ret.index = j;
	ret.found = true;
	return ret;
      }
    }
  }
  return ret;
}

ug_sel *ug_sel::clipToLayer(ug_layer *l) {
  ug_sel *ret = new ug_sel();
  ug_wire *cWire;

  for (int i = 0; i < numWires(); i++)
    if ((cWire = getWire(i))->myLayer == l)
      ret->add(cWire);

  return ret;
}

ug_sel *ug_sel::clipAboveSize(double x) {
  ug_sel *ret = new ug_sel();
  ug_wire *cWire;

  for (int i = 0; i < numWires(); i++)
    if ((cWire = getWire(i))->area() >= x || cWire->area() <= -x)
      ret->add(cWire);

  return ret;
}

void ug_sel::add(ug_wire *w) {
  int i;
  wireList.add(w);

  for (i = 0; i < w->numVerts(); i++) {
    vertList.add(w->getVert(i));
  }

  for (i = 0; i < w->numVerts() - 1 + w->isFace; i++)
    //edgeList.add(w->getEdge(i));
    edgeList.push_back(w->getEdge(i));
}

void ug_sel::merge(ug_sel *m) {
  for(int i = 0; i < m->numWires(); i++)
    add(m->getWire(i));
}

void ug_sel::setLayer(ug_layer *x) {
  for(int i = 0; i < numWires(); i++)
    getWire(i)->setLayer(x);
}

/* void ug_sel::setRegion(ug_region *x) {
  for(int i = 0; i < numWires(); i++)
    getWire(i)->setRegion(x);
    } */

void ug_sel::fOrientWires(ug_vert *v, ug_wire *b) {
  ug_vertOfWire a;

  a = anyMatchingVert(v, b);
  if (a.found) {
    if (a.w->check(ORIENTED) != 1) {
      if (a.index == a.w->numVerts() - 1) a.w->flip();
      a.w->tag(ORIENTED, 1);
      fOrientWires(a.w->getVert(a.w->numVerts() - 1), a.w);
    }
  }
}

void ug_sel::bOrientWires(ug_vert *v, ug_wire *b) {
  ug_vertOfWire a;

  a = anyMatchingVert(v, b);
  if (a.found) {
    if (a.w->check(ORIENTED) != 1) {
      if (a.index == 0) a.w->flip();
      a.w->tag(ORIENTED, 1);
      bOrientWires(a.w->getVert(0), a.w);
    }
  }
}

void ug_sel::matchOrient(ug_sel *s) {
  int i, j;
  ug_wire *cWire;

  for (i = 0; i < numWires(); i++) {
    cWire = getWire(i);
    if (cWire->check(ORIENTED) != 1) {
      for (j = 0; j < s->numWires(); j++) {
	if (cWire->matchOrient(s->getWire(j))) break;
      }
    }
  }
}

bool ug_sel::covers(ug_vert *v) {
  ug_wire *cWire;

  for (int i = 0; i < numWires(); i++) {
    cWire = getWire(i);
    for (int j = 0; j < cWire->numEdges(); j++)
      if (cWire->getEdge(j)->distance(v) < TOL) return true;
  }
      
  return false;
}

void ug_sel::writeBBox(double *a, double *b, double *c, double *d) {
  double xmin = -1, xmax = -1, ymin = -1, ymax = -1, xc, yc;
  ug_vert *cVert;
  for (int i = 0; i < numVerts(); i++) {
    cVert = getVert(i);
    xc = cVert->x;
    yc = cVert->y;
    if (xc < xmin || xmin == -1) xmin = xc;
    if (yc < ymin || ymin == -1) ymin = yc;
    if (xc > xmax || xmax == -1) xmax = xc;
    if (yc > ymax || ymax == -1) ymax = yc;
  }
  *a = xmin;
  *b = xmax;
  *c = ymin;
  *d = ymax;
}

ug_vert *ug_sel::nearestVert(ug_vert *v) {
  double td, bestDistance = -1;

  ug_vert *cVert, *best = NULL;
  for (int i = 0; i < numVerts(); i++) {
    cVert = getVert(i);
    if ((td = v->distance(cVert)) < bestDistance || bestDistance == -1) {
      bestDistance = td;
      best = cVert;
    }
  }
  return best;
}

/* void ug_sel::writeIV(FILE *fp) {
  ug_vert *cVert;
  ug_wire *cWire;
  bool first = true;
  int i;

  fprintf(fp, "Coordinate3 {\n");
  fprintf(fp, "point [\n");

  for (i = 0; i < numVerts(); i++) {
    cVert = getVert(i);
    cVert->tag(COORDINDEX, i);
    if (i != 0) fprintf(fp, ", ");
    fprintf(fp, "%f %f %f", cVert->x, cVert->y, cVert->z);
  }
  fprintf(fp, "] }\n");

  fprintf(fp, "IndexedFaceSet {\n");
  fprintf(fp, "coordIndex [\n");
  for (i = 0; i < numWires(); i++) {
    cWire = getWire(i);
    if (cWire->isFace) {
      if (!first) fprintf(fp, ", ");
      first = false;
      for (int j = 0; j < cWire->numVerts(); j++)
	fprintf(fp, "%d, ", cWire->getVert(j)->check(COORDINDEX));
      fprintf(fp, "-1");
    }
  }
  fprintf(fp, "]\n");
  fprintf(fp, "}\n");
}

void ug_sel::writeVRML(FILE *fp) {
  ug_vert *cVert;
  ug_wire *cWire;
  bool first = true;
  int i;

  fprintf(fp, "IndexedFaceSet { solid FALSE\n");
  fprintf(fp, "coord Coordinate { point [\n");
  for (i = 0; i < numVerts(); i++) {
    cVert = getVert(i);
    cVert->tag(COORDINDEX, i);
    if (i != 0) fprintf(fp, ", ");
    fprintf(fp, "%f %f %f", cVert->x, cVert->y, cVert->z);
  }
  fprintf(fp, "] }\n");

  fprintf(fp, "coordIndex [\n");
  for (i = 0; i < numWires(); i++) {
    cWire = getWire(i);
    if (cWire->isFace) {
      if (!first) fprintf(fp, ", ");
      first = false;
      for (int j = 0; j < cWire->numVerts(); j++)
	fprintf(fp, "%d ", cWire->getVert(j)->check(COORDINDEX));
      fprintf(fp, "-1");
    }
  }
  fprintf(fp, "]\n");
  fprintf(fp, "}\n");
  } */

// ug_CADSpace method definitions

ug_CADSpace::ug_CADSpace() {
  highestUsedVertNum = 1;
  highestUsedWireNum = 1;

  highestUsedInstanceNum = 1;

  nullRegion = addNewRegion("");
}

ug_vert *ug_CADSpace::addNewVert(double x, double y, double z, ug_def *d) {
  ug_vert *ret;
  ug_def *useDef = d;
  if (d == NULL) useDef = nullRegion;
  ret = new ug_vert(this, x, y, z, useDef);
  vertList.add(ret);
  return ret;
}

ug_vert *ug_CADSpace::addNewVert(char *name, double x, double y, double z, ug_def *d) {
  ug_vert *ret;
  ug_def *useDef = d;
  if (d == NULL) useDef = nullRegion;
  ret = new ug_vert(this, name, x, y, z, useDef);
  vertList.add(ret);
  return ret;
}

// this mangles the name slightly, for safety
ug_vert *ug_CADSpace::addNewNamedVert(char *name,
                                      double x, double y, double z,
                                      ug_region *region)
{
  ug_vert *ret;
  ug_region *reg;
  stringstream s(stringstream::in | stringstream::out);
  
  s << "n_" << this->unusedVertNum() << "_" << name;

  reg = (region == NULL) ? nullRegion : region;
  ret = new ug_vert(this, (char*)s.str().c_str(), x, y, z, reg);

  //assert(ret != NULL);

  vertList.add(ret);
  return ret;
}

// - "flattens" the space
// - note: addSelfToCADSpace is not recursive
// - this assumes that a region's wires only references vertices
//     in that region
// - also we are ignoring inlines right now
// - flatten should really take a matrix for transformations (optional --
// -   only for recursive flatten() calls
void ug_CADSpace::flatten()
{
  list<ug_instance*>::iterator it;
  int i, j;
  ug_instance *instance;
  ug_region *reg;
  ug_vert *v, *vnew;
  ug_wire *w, *wnew;
  map <char*, ug_vert*>tempVerts;
  stringstream s;
  int i_num;

  i_num = unusedInstanceNum();

  for (it=instanceList.begin(); it != instanceList.end(); it++) {
    instance = *it;
    reg = instance->basedOn;
    
    // ignore transforms for now
    for (i=0; i < reg->numVerts(); i++) {
      v = reg->getVert(i);
      /*
      s.str(string("")); // clears the stringstream
      s << "i_" << i_num << "_" << v->get_name();
      */
      
      vnew = this->addNewVert(v->x, v->y, v->z, NULL);
      assert(vnew != NULL);
      tempVerts[v->get_name()] = vnew;
    }

    // now loop through the wires
    for (i=0; i < reg->numWires(); i++) {
      w = reg->getWire(i);
      wnew = this->addNewWire(w->myLayer, NULL);
      if (w->isFace) wnew->isFace = true;
      assert(wnew != NULL);

      for (j=0; j < w->numVerts(); j++) {
        v = w->vertList.at(j);

        vnew = tempVerts[v->get_name()];
        assert(vnew != NULL);
        wnew->pushVert(vnew);
      }
    }

    for (i=0; i < reg->numInstances(); i++) {
      // call flatten on each of these
    }

  }
}



ug_wire *ug_CADSpace::addNewWire(ug_layer *layer, ug_def *d) {
  ug_wire *ret;
  if (d == NULL) d = nullRegion;
  wireList.add(ret = new ug_wire(this, layer, d));
  return ret;
}

void ug_CADSpace::include(char *filename, ug_def *d) {
  if (d == NULL) d = nullRegion;
  //inlineList.add(new ug_inline(this, d, filename));
  inlineList.push_back(new ug_inline(this, d, filename));
}

ug_instance *ug_CADSpace::addNewInstance(ug_def *d) {
  if (d == NULL) d = nullRegion;
  ug_instance *ret;
  //instanceList.add(ret = new ug_instance(this, d));
  instanceList.push_back(ret = new ug_instance(this, d));
  return ret;
}

ug_wire *ug_CADSpace::addNewFace(ug_layer *layer, ug_vert *v1, ug_vert *v2, ug_vert *v3, ug_vert *v4, ug_def *d) {
  ug_wire *ret;
  wireList.add(ret = new ug_wire(this, layer, d));
  ret->pushVert(v1);
  ret->pushVert(v2);
  ret->pushVert(v3);
  ret->pushVert(v4);
  ret->close();
  return ret;
}

ug_wire *ug_CADSpace::newWire(char *name) {
  return new ug_wire(name);
}

ug_layer *ug_CADSpace::addNewLayer(char *name, double r, double g, double b, int c) {
  ug_layer *ret;
  layerList.add(ret = new ug_layer(name, r, g, b, c));
  return ret;
}

ug_layer *ug_CADSpace::addExistingLayer(ug_layer *layer)
{
  layerList.add(layer);
  return layer;
}

ug_region *ug_CADSpace::addNewRegion(char *name) {
  ug_region *ret;
  //regionList.add(ret = new ug_region(name));
  regionList.push_back(ret = new ug_region(name));
  return ret;
}

void ug_CADSpace::addWire(ug_wire *w, ug_layer *layer, ug_def *d) {
  if (d == NULL) d = nullRegion;
  w->home = this;
  w->setLayer(layer);
  w->myRegion = d;
  d->add(w);
  wireList.add(w);
  for (int i = 0; i < w->numEdges(); i++)
    w->getEdge(i)->home = this;
}

int ug_CADSpace::numWires() {
  return wireList.size;
}

int ug_CADSpace::numVerts() {
  return vertList.size;
}

ug_vert *ug_CADSpace::getVertByName(char *n) {
  return vertList.get_named(n);
}

ug_wire *ug_CADSpace::getWire(int i) {
  return wireList.at(i);
}

ug_vert *ug_CADSpace::getVert(int i) {
  return vertList.at(i);
}

ug_sel *ug_CADSpace::selectAll() {
  ug_sel *ret = new ug_sel();

  for (int i = 0; i < numWires(); i++)
    ret->add(getWire(i));

  return ret;
}

/**  Method:      ug_CADSpace::print_stats( void )
 **  Description: Print out hashtable statistics for vertices.
 **/
void ug_CADSpace::print_stats()
{
  vertList.print_stats();
  return;
}

// this is a slightly bad interface in my opinion
int ug_CADSpace::unusedVertNum() {
  highestUsedVertNum++;
  return highestUsedVertNum;
}

int ug_CADSpace::unusedWireNum() {
  highestUsedWireNum++;
  return highestUsedWireNum;
}

int ug_CADSpace::unusedInstanceNum() {
  highestUsedInstanceNum++;
  return highestUsedInstanceNum;
}

ug_layer *ug_CADSpace::getLayer(char *n) {
  if (n == NULL) return NULL;

  return layerList.get_named(n);

  /*
    for (int i = 0; i < layerList.size; i++)
    if (!strcmp(layerList.at(i)->name, n)) return layerList.at(i);
    return NULL;*/
}

ug_region *ug_CADSpace::getRegion(char *n) {
  /*
  if (n == NULL) return NULL;
  for (int i = 0; i < regionList.size; i++)
    if (!strcmp(regionList.at(i)->name, n)) return regionList.at(i);
  return NULL;
  */

  unsigned int i;

  if (n == NULL) return NULL;
  for (i=0;i  < regionList.size(); i++)
    if (!strcmp(regionList.at(i)->name, n)) return regionList.at(i);
  return NULL;
}



ug_layer::ug_layer(char *n, double rc, double gc, double bc, int color) {
  strcpy(name, n);
  r = rc;
  b = bc;
  g = gc;
  c = color;
}

int ug_CADSpace::numLayers() {
  return layerList.size;
}

ug_layer *ug_CADSpace::getLayerByNumber(int i) {
  return layerList.at(i);
}

int ug_CADSpace::numRegions() {
  //return regionList.size;
  return regionList.size();
}

ug_region *ug_CADSpace::getRegionByNumber(int i) {
  return regionList.at(i);
}


ug_region *ug_CADSpace::getNullRegion()
{
  return nullRegion;
}




int ug_region::numVerts() {
  return vertList.size;
}

int ug_region::numWires() {
  return wireList.size;
}

int ug_region::numInstances() {
  //return instanceList.size;
  return instanceList.size();
}

int ug_region::numInlines() {
  //return inlineList.size;
  return inlineList.size();
}

ug_instance *ug_region::getInstance(int i) {
  return instanceList.at(i);
}

ug_inline *ug_region::getInline(int i) {
  return inlineList.at(i);
}

ug_vert *ug_region::getVert(int i) {
  return vertList.at(i);
}

ug_wire *ug_region::getWire(int i) {
  return wireList.at(i);
}

void ug_region::add(ug_vert *v) {
  vertList.add(v);
}

void ug_region::add(ug_wire *w) {
  wireList.add(w);
}

void ug_region::add(ug_instance *i) {
  //instanceList.add(i);
  instanceList.push_back(i);
}

void ug_region::add(ug_inline *i) {
  //inlineList.add(i);
  inlineList.push_back(i);
}

ug_region::ug_region(char *n) {
  if (n == NULL)
    strcpy(name, "");
  else
    strcpy(name, n);
}

void ug_CADSpace::colorFirstColorless(int color) {
  for (int i = 0; i < numLayers(); i++) {
    if (getLayerByNumber(i)->c == -1) {
      getLayerByNumber(i)->c = color;
      return;
    }
  }
}

/* void ug_CADSpace::writeIV(FILE *fp) {
  int i, j, groupsize = 100;
  ug_layer *cLayer;
  ug_sel *tempSel, *layerSel, *allSel;

  fprintf(fp, "#Inventor V2.0 ascii\n");
  fprintf(fp, "Separator {\n");
  for (i = 0; i < numLayers(); i++) {
    cLayer = getLayerByNumber(i);
    fprintf(fp, "DEF layer%s Material { diffuseColor %f %f %f }\n", cLayer->name, cLayer->r, cLayer->g, cLayer->b);
  }

  allSel = selectAll();

  for (i = 0; i < numLayers(); i++) {
    //    printf("%d\n", i);
    cLayer = getLayerByNumber(i);
    layerSel = allSel->clipToLayer(cLayer);

    j = 0;
    while (j < layerSel->numWires()) {
      if (j % groupsize == 0) tempSel = new ug_sel();
      tempSel->add(layerSel->getWire(j));
      j++;
      if (j % groupsize == 0 || j == numWires()) {
	fprintf(fp, "Separator {\n");
	fprintf(fp, "USE layer%s\n", cLayer->name);
	tempSel->writeIV(fp);
	fprintf(fp, "}\n");
	delete tempSel;
      }
    }

    delete layerSel;
  }

  fprintf(fp, "}\n");

  delete allSel;
}

void ug_CADSpace::writeVRML(FILE *fp) {
  int i, j, groupsize = 100;
  ug_layer *cLayer;
  ug_sel *tempSel, *layerSel, *allSel;

  fprintf(fp, "#VRML V2.0 utf8\n");
  for (i = 0; i < numLayers(); i++) {
    cLayer = getLayerByNumber(i);
    fprintf(fp, "DEF layer%s Appearance { material Material { diffuseColor %f %f %f\nemissiveColor %f %f %f } }\n", cLayer->name, cLayer->r / 2, cLayer->g / 2, cLayer->b / 2, cLayer->r / 2, cLayer->g / 2, cLayer->b / 2);
  }

  allSel = selectAll();

  for (i = 0; i < numLayers(); i++) {
    //    printf("%d\n", i);
    cLayer = getLayerByNumber(i);
    layerSel = allSel->clipToLayer(cLayer);

    j = 0;
    while (j < layerSel->numWires()) {
      if (j % groupsize == 0) tempSel = new ug_sel();
      tempSel->add(layerSel->getWire(j));
      j++;
      if (j % groupsize == 0 || j == numWires()) {
	fprintf(fp, "Shape {\n");
	fprintf(fp, "appearance USE layer%s\n", cLayer->name);
	fprintf(fp, "geometry ");
	tempSel->writeVRML(fp);
	fprintf(fp, "}\n");
	delete tempSel;
      }
    }

    delete layerSel;
  }

  delete allSel;
  } */

// ug_instance method definitions

ug_instance::ug_instance(ug_CADSpace *where, ug_def *d) {
  home = where;
  basedOn = NULL;
  myRegion = d;
  d->add(this);
  myLayer = NULL;
  name = "";
  transform = "";
}

ug_instance::ug_instance(ug_CADSpace *where, ug_def *d, ug_region *refersTo, double sx, double sy, double sz, double rot, double tx, double ty, double tz) {
  home = where;
  myRegion = d;
  myRegion->add(this);
  basedOn = refersTo;
  easyTransform(sx, sy, sz, rot, tx, ty, tz);
}

/* void ug_instance::setRegion(ug_def *d) {
  myRegion = d;
  myRegion->add(this);
  } */

void ug_instance::setObject(ug_def *d) {
  basedOn = d;
}

void ug_instance::setLayer(ug_layer *l) {
  myLayer = l;
}

ug_region *ug_instance::getRegion() {
  return myRegion;
}

ug_region *ug_instance::getObject() {
  return basedOn;
}

ug_layer *ug_instance::getLayer() {
  return myLayer;
}

void ug_instance::makeTransform(char *x) {
  stringstream s(stringstream::in | stringstream::out);

  s << x;
  transform = s.str();
}

void ug_instance::setName(char *x) {
  stringstream s(stringstream::in | stringstream::out);

  s << x;
  transform = s.str();
}

void ug_instance::easyTransform(double sx, double sy, double sz,
                                double rot, double tx, double ty, double tz) {
  stringstream s(stringstream::in | stringstream::out);
  
  s << "-sx " << sx << " -sy " << sy << " -sz " << sz << " -rz " << rot
    << " -tx " << tx << " -ty " << ty << " -tz " << tz;
  transform = s.str();
}

const char *ug_instance::getTransform() {
  return transform.c_str();
}

const char *ug_instance::getName() {
  return name.c_str();
}

// ug_inline method definitions

ug_inline::ug_inline(ug_CADSpace *where, ug_def *d, char *x) {
  home = where;
  myRegion = d;
  myRegion->add(this);
  strcpy(filename, x);
  strcpy(name, "");
}

/* void ug_inline::setRegion(ug_def *d) {
  if (r == NULL)
    myRegion = home->nullRegion;
  else
    myRegion = r;
  myRegion->add(this);
  } */

void ug_inline::setFilename(char *x) {
  strcpy(filename, x);
}

ug_region *ug_inline::getRegion() {
  return myRegion;
}

char *ug_inline::getFilename() {
  return filename;
}

// floorModel method definitions

/* floorModel::floorModel() {
  solids = new ug_sel();
  inWalls = new ug_sel();
  outWalls = new ug_sel();
  floors = new ug_sel();
  verbatim = new ug_sel();
  ceilings = new ug_sel();
  actualDoors = new ug_sel();
  shafts = new ug_sel();
  skinSel = new ug_sel();
}

floorModel::~floorModel() {
  delete solids, inWalls, outWalls, floors, verbatim, ceilings, actualDoors, skinSel;
}

ug_sel *floorModel::extrudeLowLOD(ug_CADSpace *where, double height) {
  ug_sel *tempSel;
  ug_sel *ret = new ug_sel();
  int i;

  for (i = 0; i < skinSel->numWires(); i++) {
    tempSel = skinSel->getWire(i)->extrude(0, height);
    ret->merge(tempSel);
    delete tempSel;
  }

  return ret;
}

ug_sel *floorModel::extrude(ug_CADSpace *where) {
  ug_sel *ret, *tempSel;
  ug_wire *cWire, *nWire;
  ug_vert *vf[4][6];
  int i, j, k;
  double hfield[6];
  portal *tp;

  ret= new ug_sel();

  printf("Extruding floor\n");

  printf("Extruding structural elements\n");

  for(i = 0; i < solids->numWires(); i++) {
    tempSel = solids->getWire(i)->extrude(zBase, zTop);
    if (solids->getWire(i)->check(ORIENTED) == 1)
      tempSel->setLayer(where->getLayer("OSTRUCTURAL"));
    else
      tempSel->setLayer(where->getLayer("STRUCTURAL"));
    ret->merge(tempSel);
    delete tempSel;
  }

  for (i = 0; i < verbatim->numWires(); i++)
    ret->add(verbatim->getWire(i));

  printf("Extruding floors and ceilings\n");

  for (i = 0; i < floors->numWires(); i++) {
    nWire = floors->getWire(i)->dup(0, 0, zFloor);
    nWire->setLayer(where->getLayer("FLOOR"));
    ret->add(nWire);
  }

  for (i = 0; i < shafts->numWires(); i++) {
    nWire = shafts->getWire(i)->dup(0, 0, zFloor);
    nWire->setLayer(where->getLayer("SHAFT"));
    ret->add(nWire);
  }

  for (i = 0; i < ceilings->numWires(); i++) {
    nWire = floors->getWire(i)->dup(0, 0, zCeiling);
    nWire->setLayer(where->getLayer("CEILING"));
    nWire->flip();
    ret->add(nWire);
  }

  printf("Extruding walls\n");

  for(i = 0; i < inWalls->numWires(); i++) {
    tempSel = inWalls->getWire(i)->extrude(zFloor, zCeiling);
    if (inWalls->getWire(i)->check(ORIENTED) == 1)
      tempSel->setLayer(where->getLayer("OWALL"));
    else
      tempSel->setLayer(where->getLayer("WALL"));
    ret->merge(tempSel);
    delete tempSel;
  }

  printf("Extruding doors and windows\n");

  for (i = 0; i < portals.size; i++) {
    tp = portals.at(i);

    hfield[0] = zFloor;
    hfield[3] = hfield[4] = zCeiling;

    if (tp->outside[0]) {
      hfield[3] = zTop;
      hfield[0] = zBase;
    }
    if (tp->outside[1]) {
      hfield[4] = zTop;
      hfield[5] = zBase;
    }
    
    if (tp->isWindow) {
      hfield[1] = zWindowB;
      hfield[2] = zWindowT;
    } else {
      hfield[1] = zFloor;
      hfield[2] = zDoorT;
    }

    for (j = 0; j < 4; j++)
      for (k = 0; k < 6; k++)
	vf[j][k] = tp->jambs[(j & 2) / 2]->getVert(j & 1)->dup(0, 0, hfield[k]);

    ret->add(where->addNewFace(where->getLayer("WINDOWFRAME"), vf[0][1], vf[1][1], vf[2][1], vf[3][1]));
    ret->add(where->addNewFace(where->getLayer("WINDOWFRAME"), vf[3][2], vf[2][2], vf[1][2], vf[0][2]));
    ret->add(where->addNewFace(where->getLayer("WINDOWFRAME"), vf[0][1], vf[0][2], vf[1][2], vf[1][1]));
    ret->add(where->addNewFace(where->getLayer("WINDOWFRAME"), vf[2][1], vf[2][2], vf[3][2], vf[3][1]));
    ret->add(where->addNewFace(where->getLayer("OWALL"), vf[0][2], vf[0][3], vf[3][3], vf[3][2]));
    ret->add(where->addNewFace(where->getLayer("OWALL"), vf[2][2], vf[2][4], vf[1][4], vf[1][2]));
    if (tp->isWindow) {
      ret->add(where->addNewFace(where->getLayer("OWALL"), vf[0][0], vf[0][1], vf[3][1], vf[3][0]));
      ret->add(where->addNewFace(where->getLayer("OWALL"), vf[2][5], vf[2][1], vf[1][1], vf[1][5]));
    }
  }

  return ret;
}

void floorModel::interpret(ug_sel *plan, ug_CADSpace *home) {
  int i, j, bi;
  ug_wire *c, *w1, *closestWire;
  char *ln;
  double best = 0;
  ug_edge *e1;
  ug_vert *v1, *v2;
  ug_vertOfWire vw1;
  ug_sel *wallSel, *doorSel, *iWallSel, *windowSel, *solidSel, *floorSel;
  portal *tp;

  printf("Interpreting plan\n");

  printf("Dividing layers\n");

  iWallSel = plan->clipToLayer(home->getLayer("A_WALL"));
  wallSel = plan->clipToLayer(home->getLayer("A_WALL_CORE"));
  solidSel = plan->clipToLayer(home->getLayer("A_WALL_CORE"));
  wallSel->merge(iWallSel);
  doorSel = plan->clipToLayer(home->getLayer("A_DOOR_JAMB"));
  windowSel = plan->clipToLayer(home->getLayer("A_WINDOW_JAMB"));
  floorSel = plan->clipToLayer(home->getLayer("A_AREA_FICM"));
  skinSel = plan->clipToLayer(home->getLayer("A_AREA_FICM_GROS"));

  printf("Orienting walls using FICM layers\n");

  for (i = 0; i < floorSel->numWires(); i++) {
    c = floorSel->getWire(i);
    if (c->area() > 0) c->flip();
    floors->add(c);
    ceilings->add(c);
  }

  solidSel->matchOrient(skinSel);
  //  iWallSel->matchOrient(floorSel);

  printf("Beginning first pass through wires\n");

  for (i = 0; i < plan->numWires(); i++) {
    c = plan->getWire(i);
    ln = c->myLayer->name;

    if (!strcmp(ln, "A_WALL_CORE")) {
      if (c->area() < 0) c->flip();
      solids->add(c);
    } else if (!strcmp(ln, "A_WALL")) {
      c->flip();
    } else if (!strcmp(ln, "A_AREA_FICM_SHFT")) {
      if (c->area() > 0) c->flip();
      shafts->add(c);
    } else if (!strcmp(ln, "A_AREA_FICM_GROS")) {
      verbatim->add(c);
    } else if (!strcmp(ln, "A_DOOR_JAMB") || !strcmp(ln, "A_WINDOW_JAMB")) {
      v1 = c->getVert(0);
      vw1 = wallSel->anyMatchingVert(v1, NULL);
      if (vw1.found) {
	if (v1->match(vw1.w->getVert(0)) && !c->getEdge(0)->isAbove(vw1.w->getVert(1)))
	  c->flip();
	else if (v1->match(vw1.w->getVert(vw1.w->numVerts() - 1)) && !c->getEdge(0)->isAbove(vw1.w->getVert(vw1.w->numVerts() - 2)))
	  c->flip();
	iWallSel->fOrientWires(c->getVert(1), c);
	iWallSel->bOrientWires(c->getVert(0), c);
      }
    }
  }      

  printf("Adding walls\n");

  for (i = 0; i < iWallSel->numWires(); i++) {
    inWalls->add(iWallSel->getWire(i));
  }

  printf("Pairing off door jambs\n");

  for (i = 0; i < doorSel->numWires(); i++) {
    c = doorSel->getWire(i);
    best = 0;
    closestWire = NULL;

    if (c->check(PAIRED) == 0) {
      for (j = i + 1; j < doorSel->numWires(); j++) {
	w1 = doorSel->getWire(j);
	e1 = w1->getEdge(0);
	if (w1->check(PAIRED) == 0 && !e1->isAbove(c->getVert(0)) && e1->isParallel(c->getEdge(0)) && e1->overlaps(c->getEdge(0)))
	  if (best == 0 || e1->distance(c->getEdge(0)) < best) {
	    best = e1->distance(c->getEdge(0));
	    bi = j;
	    closestWire = w1;
	  }
      }
      if (closestWire != NULL) {
	c->tag(PAIRED, 1);
	closestWire->tag(PAIRED, 1);
	portals.add(new portal(c, closestWire, false, false, false));
      }
    }
  }

  printf("Pairing off window jambs\n");

  for (i = 0; i < windowSel->numWires(); i++) {
    c = windowSel->getWire(i);
    best = 0;
    closestWire = NULL;

    if (c->check(PAIRED) == 0) {
      for (j = i + 1; j < windowSel->numWires(); j++) {
	w1 = windowSel->getWire(j);
	e1 = w1->getEdge(0);
	if (w1->check(PAIRED) == 0 && !e1->isAbove(c->getVert(0)) && e1->isParallel(c->getEdge(0)) && e1->overlaps(c->getEdge(0)))
	  if (best == 0 || e1->distance(c->getEdge(0)) < best) {
	    best = e1->distance(c->getEdge(0));
	    bi = j;
	    closestWire = w1;
	  }
      }
      if (closestWire != NULL) {
	c->tag(PAIRED, 1);
	closestWire->tag(PAIRED, 1);
	portals.add(new portal(c, closestWire, true, skinSel->covers(c->getVert(0)) || skinSel->covers(closestWire->getVert(1)), skinSel->covers(closestWire->getVert(0)) || skinSel->covers(c->getVert(1))));
      }
    }
    } */

  /*
  for (i = 0; i < plan->numWires(); i++) {
    c = plan->getWire(i);
    if (!strcmp(c->myLayer->name, "A_DOOR")) {
      w1 = new wire(c->home);
      if (c->getEdge(0)->length() > c->getEdge(c->numVerts() - 2)->length()) {
	w1->add(c->getVert(0));
	w1->add(c->getVert(1));
      } else {
	w1->add(c->getVert(c->numVerts() - 2));
	w1->add(c->getVert(c->numVerts() - 1));
      }
      actualDoors->add(w1);
      verbatim->add(c);
    } else if (!strcmp(c->myLayer->name, "A_DOOR_JAMB") || !strcmp(c->myLayer->name, "A_WINDOW_JAMB")) {
      }
    }
    //    printf("%s ", c->myLayer->name);
  }

}
*/

/*************************************************************************/
/**                        FUNCTION DEFINITIONS                         **/
/*************************************************************************/

/**  Function:    init_datastructures()
 **  Description: Fill the hashing algorithm table with random numbers.
 **/
void init_datastructures()
{
  int i, j;               /** Iterators. **/

  srandom(time(0));

  for(i = 0; i < ALG_ARRAY_SIZE_1; ++i){
    for(j = 0; j < ALG_ARRAY_SIZE_2; ++j){
      alg_table[i][j] = (unsigned int) random();
    }
  }

  return;
}

// Methods for GPC support

void ug_sel::writeGPC(gpc_polygon *p) {
  p->num_contours = numWires();
  p->hole = (int *) malloc(p->num_contours * sizeof(int));
  p->contour = (gpc_vertex_list *) malloc(p->num_contours * sizeof(gpc_vertex_list));
  for (int c = 0; c < p->num_contours; c++)
    getWire(c)->writeGPC(&(p->contour[c]));
}

void ug_wire::writeGPC(gpc_vertex_list *c) {
  c->num_vertices = numVerts();
  c->vertex = (gpc_vertex *) malloc(c->num_vertices * sizeof(gpc_vertex));
  for (int v = 0; v < c->num_vertices; v++) {
    c->vertex[v].x = getVert(v)->x;
    c->vertex[v].y = getVert(v)->y;
  }
}

ug_sel::ug_sel(gpc_polygon *p, ug_CADSpace *where, ug_layer *layer, ug_def *d, double z) {
  ug_wire *nWire;

  for (int i = 0; i < p->num_contours; i++) {
    nWire = where->addNewWire(&(p->contour[i]), layer, d, z);
    if (nWire->area() > 0) nWire->flip();
    add(nWire);
  }
}

ug_wire *ug_CADSpace::addNewWire(gpc_vertex_list *c, ug_layer *layer, ug_def *d, double z) {
  ug_wire *ret = addNewWire(layer, d);

  for (int i = 0; i < c->num_vertices; i++)
    ret->pushVert(addNewVert(c->vertex[i].x, c->vertex[i].y, z, d));
  
  ret->close();

  return ret;
}

/** these two functions are used to determine whether edges intersect
    with each other */

int turn(ug_vert v1, ug_vert v2, ug_vert v3) {
  if (v1.x * (v2.y - v3.y) + 
      v2.x * (v3.y - v1.y) + 
      v3.x * (v1.y - v2.y)
      > 0)
    return 1;
  else
    return -1;
}

bool intersect(const ug_edge* edge1, const ug_edge* edge2) {
  ug_vert * a1 = edge1->aVerts[0];
  ug_vert * a2 = edge1->aVerts[1];
  ug_vert * b1 = edge2->aVerts[0];
  ug_vert * b2 = edge2->aVerts[1];
  return turn (*a1, *a2, *b1) == -turn (*a1, *a2, *b2) && 
           turn (*b1, *b2, *a1) == -turn (*b1, *b2, *a2); 
}

/** determine if two wires overlap horizontally.  useful for vertical
    adjacency determinations.
*/

bool intersect(ug_wire* wire1, ug_wire* wire2) {
  int wire1Verts = wire1->numVerts();
  int wire2Verts = wire2->numVerts();
  int j;

  //check for vertices inside one another
  for (j = 0; j < wire2Verts; j++) {
    ug_vert* wire2Vert = wire2->getVert(j);
    if (inside(wire2Vert, wire1))
      return true;
  }
  for (j = 0; j < wire1Verts; j++) {
    ug_vert* wire1Vert = wire1->getVert(j);
    if (inside(wire1Vert, wire2))
      return true;
  }
  //now, check for edges which intersect with one another
  for (int i = 0; i < wire1->numEdges(); i++) {
    ug_edge* nextEdge1 = wire1->getEdge(i);
    for (int j = 0; j < wire2->numEdges(); j++) {
      ug_edge* nextEdge2 = wire2->getEdge(j);
      if (intersect(nextEdge1, nextEdge2))
	return true;
    }
  }
  //spaces must not intersect
  return false;
}

bool inside(ug_vert* vertex, ug_wire* wire) {
  /* find a point outside the wire */
  ug_CADSpace * cadSpace = new ug_CADSpace();
  
  int numRoomVerts = wire->numVerts();
  double maxX = -MAXFLOAT;
  for (int j = 0; j < numRoomVerts; j++) {
    ug_vert* nextVert = wire->getVert(j);
    if (nextVert->x > maxX)
      maxX = nextVert->x;
  }
  //it doesn't matter what the y coordinate is, since the X
  //coordinate is guaranteed outside the polyline
  ug_vert* outsideVert = cadSpace->addNewVert("outsideVert", 
					      maxX + 1, 0.0, 0.0, vertex->myRegion);
  assert(outsideVert->x == maxX + 1 && outsideVert->y == 0.0);
  
  //make an edge
  ug_wire * wire1 = cadSpace->addNewWire(NULL,vertex->myRegion);
  wire1->pushVert(outsideVert);
  wire1->pushVert(vertex);
  ug_edge* testEdge = wire1->getEdge(0);
  
  //count intersecting room edges
  int intersectionCount = 0;
  for (int k = 0; k < wire->numEdges(); k++) {
    ug_edge* nextEdge = wire->getEdge(k);
    if (intersect(nextEdge, testEdge))
      intersectionCount++;
  }
  //even or odd?
  if ((intersectionCount%2) == 1)
    return true;
  else
    return false;
}
