#include <string>
#include <sstream>
#include <stdlib.h>
#include <assert.h>
#include "StringPack.h"

#define VALID_UG_CHAR(x) ( ( (x)>='A' && (x)<='Z' ) ||          \
                           ( (x)>='a' && (x)<='z' ) ||          \
                           ( (x)>='0' && (x)<='9' ) ||          \
                           ( (x) == ':' ) ||                    \
                           ( (x) == '_' ) ||                    \
                           ( (x) == '#' ) )

StringPack::StringPack()
  : str_rep(""),
    changed(false)
{

}

StringPack::~StringPack()
{

}

void StringPack::addString(string key, string &value)
{
  // should check for a previous value!

  pairs[key] = value;

  changed = true;
}

void StringPack::addString(string key, const string &value)
{
  // should check for a previous value!

  pairs[key] = value;

  changed = true;
}

void StringPack::addInt(string key, int value)
{
  stringstream s;

  s << value;
  pairs[key] = s.str();

  changed = true;
}


void StringPack::addDouble(string key, double value)
{
  stringstream s;

  s << value;
  pairs[key] = s.str();

  changed = true;
}


const string StringPack::getString(string key)
{
  // might want to assert that the value is non-empty

  return pairs[key];
}

int StringPack::getInt(string key)
{
  char *cp, *endp;
  int ret;

  // again -- assert that non-empty
  
  cp = (char*) pairs[key].c_str();
  ret = strtol(cp, &endp, 10);

  assert( !( ret==0 && endp==cp ) );

  return ret;
}

double StringPack::getDouble(string key)
{
  char *cp, *endp;
  double ret;

  // again -- assert that non-empty

  cp = (char*) pairs[key].c_str();
  ret = strtod(cp, &endp);

  assert( !( ret==0.0 && endp==cp ) );

  return ret;
}




const string StringPack::getString(char *key)
{
  // might want to assert that the value is non-empty

  return pairs[string(key)];
}

int StringPack::getInt(char *key)
{
  return this->getInt(string(key));
}

double StringPack::getDouble(char *key)
{
  return this->getDouble(string(key));
}




/*
  - right now, this does not check for characters invalid to UG --
    this check may never be done (here)
  - "#s" is a separator; "##" stands for '#'
*/
string StringPack::serialize()
{
  map<string, string>::iterator it;
  stringstream s(stringstream::in | stringstream::out);
  char *cp, *sp;
  int i;
  bool first = true;
  unsigned char c;
  char oa, ob, oc;   // for octal representation of an invalid char

  if (changed) {
    for (it = pairs.begin(); it != pairs.end(); it++) {

      if (!first) s << "#s";

      cp = (char*) it->first.c_str();
      for (i=0, sp=cp; cp[i] != '\0'; i++) {
        
        if (cp[i] == '#') {
          cp[i] = '\0';
          s << sp;
          sp = cp + i+1;
          s << "##";

        } else if (!VALID_UG_CHAR(cp[i])) {
          c = (unsigned char) cp[i];
          oc = (c & 0x07) + '0';
          c >>= 3;
          ob = (c & 0x07) + '0';
          c >>= 3;
          oa = (c & 0x07) + '0';

          cp[i] = '\0';
          s << sp;
          s << "#x" << oa << ob << oc;

          sp = cp + i+1;

        }
      }
      s << sp;

      s << "#s";

      cp = (char*) it->second.c_str();
      for (i=0, sp=cp; cp[i] != '\0'; i++) {
        
        if (cp[i] == '#') {
          cp[i] = '\0';
          s << sp;
          sp = cp + i+1;
          s << "##";
        } else if (!VALID_UG_CHAR(cp[i])) {
          c = (unsigned char) cp[i];
          oc = (c & 0x07) + '0';
          c >>= 3;
          ob = (c & 0x07) + '0';
          c >>= 3;
          oa = (c & 0x07) + '0';
          
          cp[i] = '\0';
          s << sp;
          s << "#x" << oa << ob << oc;

          sp = cp + i+1;
        }


      }
      s << sp;
      
      first = false;
    }
    str_rep = s.str();

    changed = false;
  }

  return str_rep;
}

/*
  - this completely replaces whatever was being held before
*/
void StringPack::deserialize(string &encstr)
{
  char *cp, *sp;
  int i;
  int parity;
  //char type;
  stringstream sstr(stringstream::in | stringstream::out);
  string key;
  char octd;
  unsigned char octc;

  str_rep = encstr;
  pairs.clear();

  parity = 0;
  //type = '\0'; // i.e., no type

  cp = (char*) str_rep.c_str();
  for (i=0, sp=cp; cp[i] != '\0'; i++) {
    
    if (cp[i] == '#') {
      switch (cp[i+1]) {
        
      case '#':
        cp[i] = '\0';
        sstr << sp << '#';
        sp = cp + i+2;
        i++;
        break;

        // separator between pairs
      case 's':
        //assert(parity == 0);
        parity ^= 1;
        cp[i] = '\0';
        sstr << sp;
        //key = sstr.str();
        if (parity == 1) key = sstr.str();
        else pairs[key] = sstr.str();
        //printf("got the string: \"%s\"\n", sstr.str().c_str());
        sstr.str(string(""));
        sp = cp + i+2;
        //type = '\0';
        break;

      case 'x':
        // this should really be an exception (all of those immediately below)
        assert(cp[i+2]!='\0' && cp[i+3]!='\0' && cp[i+4]!='\0');
        
        octc = 0;

        octd = cp[i+2];
        assert(octd>='0' && octd<='9');
        octc += octd - '0';

        octd = cp[i+3];
        assert(octd>='0' && octd<='9');
        octc = (octc<<3) + octd - '0';

        octd = cp[i+4];
        assert(octd>='0' && octd<='9');
        octc = (octc<<3) + octd - '0';

        cp[i] = '\0';
        sstr << sp << octc;
        sp = cp + i+5;
        i += 4;

        break;

        /*
        // the characters below are type indicators for the values of pairs

        // this indicates a string
      case 'S':
        assert(parity == 1);
        parity ^= 1;
        type = 'S';
        break;

        // this indicates an int
      case 'i':
        assert(parity == 1);
        parity ^= 1;
        break;

        // this indicates a double
      case 'd':
        assert(parity == 1);
        parity ^= 1;
        break;
        */

        // if any other char is used, just interpret it as that character
      default:
        cp[i] = '\0';
        sstr << sp;
        sstr << cp[i+1];
        sp = cp + i+2;
        i++;
        break;

      }
    }


  }
  if (*sp != '\0') sstr << sp;
  assert(parity == 1);
  pairs[key] = sstr.str();


  changed = false;
  
}
