/**  File:        support.cpp
 **  Package:     none
 **  Created by:  markan, wml
 **  Modified by: markan
 **  Description: Template class implementations
 **
 **  It should not be necessary to compile this file directly;
 **  CC takes care of it because it assumes that this file
 **  implements the templates declared in support.h.  Just
 **  include that file to get the functionality of this one.
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "support.h"

/****************************************************************************/
/**                           METHOD DEFINITIONS                           **/
/****************************************************************************/

// ** node<T> template class method definitions

#ifdef ZERO

template <class T> node<T>::node() {
  prev = next = NULL;
}

// ** list<T> template class method definitions

template <class T> list<T>::list() {
  current = first = last = NULL;
  currentIndex = 0;
  size = 0;
}

template <class T> bool list<T>::contains(T data) {
  node<T> *check = first;
  for(int j = 0; j < size; j++) {
    if (check->data == data) return true;
    check = check->next;
  }
  return false;
}

template <class T> int list<T>::find(T data) {
  for(int i = 0; i < size; i++)
    if (at(i) == data) return i;
  return size;
}

/* template <class T> int list<T>::size() {
  return n;
  } */

template <class T> T list<T>::at(int i) {
  if (i == currentIndex + 1) {
    currentIndex++;
    current = current->next;
  } else if (i == currentIndex - 1) {
    currentIndex--;
    current = current->prev;
  } else if (i != currentIndex) {
    currentIndex = i;
    current = first;
    for(int j = 0; j < i; j++) current = current->next;
  }
  return current->data;
}

template <class T> void list<T>::add(T data) {
  size++;
  if (first == NULL) {
    current = first = last = new node<T>();
    first->data = data;
  } else {
    last->next = new node<T>();
    last->next->prev = last;
    last = last->next;
    last->data = data;
  }
}

template <class T> void list<T>::set(int i, T data) {
  if (i == currentIndex + 1) {
    currentIndex++;
    current = current->next;
  } else if (i == currentIndex - 1) {
    currentIndex--;
    current = current->prev;
  } else if (i != currentIndex) {
    currentIndex = i;
    current = first;
    for(int j = 0; j < i; j++) current = current->next;
  }
  current->data = data;
}

#endif

// ** table_node - Nodes for a hashtable.

/**  Method:      table_node<T>::table_node( void )
 **  Description: Default constructor.  Set everything to NULL.
 **/
template <class T>
table_node<T>::table_node()
{
  next = prev = next_ht = NULL;
  data = NULL;
}

// ** h_table - Hash table.

/**  Method:      *h_table<T>::new_node( void )
 **  Description: Return a new (unused) node.
 **/
template <class T>
table_node<T> *h_table<T>::new_node() { return new table_node<T>; }

/**  Method:      h_table<T>::hash_str(*str)
 **  Description: Hash the given string *str to a bucket.  This works by using a table
 **    of random numbers.  There is one random number for every letter in every position
 **    up to the ALG_ARRAY_SIZE_2 position.  It takes all of the values of the name and
 **    XOR's them together to get the hash value.  It then AND's this result with one
 **    less than the size of the table (a power of 2) to get the appropriate bucket.  I
 **    was simply adding the integer values of each character together and doing the AND
 **    step, which is substancially faster, but only hashes to the first 128 buckets (or
 **    so) for short names.  As this works, for any number of entries up to the maximum
 **    unsigned int value, you are pretty much guaranteed to get O(C) access time.
 **/
template <class T>
unsigned int h_table<T>::hash_str(char *str)
{
  unsigned int loc;             /** Location to which the object will be hashed. **/
  int i;                        /** Iterator. **/

  assert(str != NULL);

  loc = 0;
  for(i = 0; str[i] != '\0' && i < ALG_ARRAY_SIZE_2; ++i)
    loc = (loc ^ alg_table[(int) str[i]][i]);

  /** Take the low order bits that are the important ones. **/
  loc = (loc & (table_length - 1));

  return loc;
}

/**  Method:      h_table<T>::rehash( void )
 **  Description: Resize the table and rehash all of the elements, therein.  This helps
 **    assure that the random access time is never linear.
 **/
template <class T>
void h_table<T>::rehash()
{
  unsigned int old_t_size;         /** Old table size. **/
  unsigned int loc;                /** New location to hash an item. **/
  table_node<T> **new_buckets;     /** New array of buckets for the table. **/
  table_node<T> *p, *q;            /** Iterators. **/
  unsigned int i;                           /** Iterator. **/

  /** Resize the array of buckets. **/
  old_t_size = table_length;
  for( ; table_length < size; table_length *= 2) {}
    //    fprintf(stderr, "  Resizing HT to %d (size = %d)\n", table_length, size);
  new_buckets = (table_node<T>**) calloc(table_length, sizeof(table_node<T>*));

  /** Rehash all of the data into the new set of buckets. **/
  for(i = 0; i < old_t_size; ++i){
    for(p = buckets[i]; p != NULL; p = q){
      q = p->next_ht;

      assert(p->data->get_name() != NULL);

      loc = hash_str(p->data->get_name());

      assert(loc >= 0);
      assert(loc < table_length);

      p->next_ht = new_buckets[loc];
      new_buckets[loc] = p;
    }
  }

  /** Free up the extra space. **/
  free(buckets);
  buckets = new_buckets;

  return;
}

/**  Method:      h_table<T>::h_table( void )
 **  Description: Default constructor.  Set up the table for accessing.
 **/
template <class T>
h_table<T>::h_table()
{
  /** List stuff. **/
  current = first = NULL;
  currentIndex = 0;
  size = 0;

  /** Hash table stuff. **/
  table_length = DEFAULT_TABLE_SIZE;
  buckets = (table_node<T>**) calloc(table_length, sizeof(table_node<T>*));
}

/**  Method:      h_table<T>::~h_table( void )
 **  Description: Destructor.  Free up the space used by the table and the nodes.
 **/
template <class T>
h_table<T>::~h_table()
{
  free(buckets);
}

/**  Method:      h_table<T>::contains(data)
 **  Description: Return true if the element, data, exists within the hash table.
 **    Otherwise, return false.  This is a very slow operation.
 **/
template <class T>
bool h_table<T>::contains(T* data)
{
  table_node<T> *check = first;      /** Iterator. **/

  for(int j = 0; j < size; j++) {
    if (check->data == data) return true;
    check = check->next;
  }
  return false;
}

/**  Method:      h_table<T>::find(data)
 **  Description: Return the location (in list format) of the desired piece of data.
 **    If it doesn't exist, return a number one past the end of the list.  This is a
 **    very slow operation.  Please don't use it.
 **/
template <class T>
int h_table<T>::find(T* data)
{
  for(int i = 0; i < size; i++)
    if (at(i) == data) return i;
  return size;
}

/**  Method:      h_table<T>::size( void )
 **  Description: Return the number of elements hashed into the table.
 **/
/* template <class T>
   int h_table<T>::size(){ return n; } */

/**  Method:      *h_table<T>::at(i)
 **  Description: Return the element located (in a list-sort of way) at the index, i.
 **    CAUTION: Don't use this function unless you're iterating through the list one
 **             element at a time.  If you're using this for random access, it will be
 **             extremely slow due to its O(n) nature.
 **/
template <class T>
T *h_table<T>::at(int i)
{
  if (i == currentIndex + 1) {
    currentIndex++;
    current = current->next;
  } else if (i == currentIndex - 1) {
    currentIndex--;
    current = current->prev;
  } else if (i != currentIndex) {
    currentIndex = i;
    current = first;
    for(int j = 0; j < i; j++) current = current->next;
  }
  return current->data;
}

/**  Method:      h_table<T>::add(data)
 **  Description: Add a node containing 'data.'
 **/
template <class T>
void h_table<T>::add(T *data)
{
  table_node<T> *in;            /** Node to insert. **/
  unsigned int loc;             /** Bucket in which to insert the node. **/

  ++size;

  if(size > table_length * 8) rehash();

  /** Insert it into the list. **/
  if (first == NULL) {
    in = new_node();
    current = first = in;
    first->data = data;

    first->next = first;
    first->prev = first;
  } else {
    in = new_node();
    in->data = data;

    in->next          = first;
    in->prev          = first->prev;
    first->prev->next = in;
    first->prev       = in;
  }

  /** Insert it into the table. **/
  loc = hash_str(data->get_name());
  assert(loc >= 0);
  assert(loc < table_length);

  in->next_ht  = buckets[loc];
  buckets[loc] = in;

  return;
}

/*
  removes the element from the hash table, if it is there;
  returns true if it was there, false otherwise
*/
/*
template <class T>
bool h_table<T>::remove(T *data)
{
  unsigned int loc;
  table_node<T> *p, *pn;
  char *str;

  str = data->get_name();
  loc = hash_str(str);

  assert(loc >= 0 && loc < table_length);

  if (!strcmp(str, buckets[loc]->data->get_name())) {
    // the first node in the bucket is the one to delete

    p = buckets[loc];
    buckets[loc] = p->next_ht;
    if (first == p) {
      first = NULL;
      current = NULL;
      currentIndex = 0;
      delete p;
      return true;
    }
    p->prev->next = p->next;
    p->next->prev = p->prev;
    delete p;
    return true;
  }

  p = buckets[loc];
  for (pn = p->next_ht; pn != NULL; p=pn, pn=pn->next_ht) {
    if (!strcmp(str, pn->data->get_name())) {
      // pn is the one: delete it

      p->next_ht = pn->next_ht;
      assert(first != pn);
      if (current == pn) {
        current = pn->next;
        if (current == first) currentIndex = 0;
        else currentIndex++;
      }

      pn->prev->next = pn->next;
      pn->next->prev = pn->prev;
    }
  }

}
*/


/**  Method:      h_table<T>::set(i, data)
 **  Description: Set the contents of the node at index, i, to the value, data.
 **/
template <class T>
void h_table<T>::set(int i, T* data)
{
  if (i == currentIndex + 1) {
    currentIndex++;
    current = current->next;
  } else if (i == currentIndex - 1) {
    currentIndex--;
    current = current->prev;
  } else if (i != currentIndex) {
    currentIndex = i;
    current = first;
    for(int j = 0; j < i; j++) current = current->next;
  }
  current->data = data;

  return;
}

/**  Method:      *h_table<T>::get_named(*str)
 **  Description: Get the vertex with the name, str.  If no such vertex exists, return
 **    NULL.
 **/
template <class T>
T *h_table<T>::get_named(char *str)
{
  unsigned int loc;             /** Where the node resides if it exists. **/
  table_node<T> *p;             /** Iterator. **/
  int len = 0;

  loc = hash_str(str);

  assert(loc >= 0);
  assert(loc < table_length);

  for(p = buckets[loc]; p != NULL; p = p->next_ht){
    if(!strcmp(str, p->data->get_name())) return p->data;
    len++;
  }

  if ( len > 16 )
    fprintf ( stderr, "listlen was %d\n", len);

  return NULL;
}

/**  Method:      h_table<T>::print_stats( void )
 **  Description: Print statistics for each bucket.
 **/
template <class T>
void h_table<T>::print_stats()
{
  int empty_buckets;            /** Number of empty buckets. **/
  unsigned int i;                        /** Iterator. **/

  empty_buckets = 0;
  for(i = 0; i < table_length; ++i) if(buckets[i] == NULL) ++empty_buckets;

  fprintf(stderr, "  Buckets = %d, Items = %d\n", table_length, size);
  fprintf(stderr, "  Empty Buckets = %d\n", empty_buckets);

  return;
}

