/**  File:        support.h
 **  Package:     none
 **  Created by:  markan, wml
 **  Modified by: markan
 **  Description: Template classes for lists, hashes, and the like
 **/

#ifndef _SUPPORT_H_
#define _SUPPORT_H_


/** Constants for the hash table algorithm
    The first should always be a power of 2 **/
#define DEFAULT_TABLE_SIZE 128
#define ALG_ARRAY_SIZE_1 128
#define ALG_ARRAY_SIZE_2 16


/** Global randomized array for use in hash algorithm **/

// I don't think these directives are necessary but just in case...

#ifndef ALG_TABLE_INT
#define ALG_TABLE_INT
extern unsigned int alg_table[ALG_ARRAY_SIZE_1][ALG_ARRAY_SIZE_2];
#endif


/** Supporting data structures **/

#ifdef ZERO

template <class T> struct node {
  T data;
  node<T> *prev, *next;

  node();
};

template <class T> struct list {
 public:
  int size, currentIndex;
  node<T> *first, *last, *current;

  list();
  bool contains(T data);
  int find(T data);
  T at(int i);
  void add(T data);
  void set(int i, T data);
};

#endif

/**  struct:      table_node
 **  Description: Nodes used by a hash table.  They can be accessed linearly (like a
 **    list) or randomly (like a hash table).
 **/
template <class T>
struct table_node{
  T *data;                       /** Contents of this node. **/
  table_node<T> *prev, *next;    /** Pointers for access like a list. **/
  table_node<T> *next_ht;        /** Pointer to the next element in the bucket. **/

  /** Structure Methods: see implementations for details. **/

  table_node();
};

/**  class:       h_table
 **  Description: Hash table implementation.  Can be accessed like a linked list.  Any
 **    class T that is used must be a pointer, and must have a method called get_name()
 **    that returns a string of characters for hashing purposes.
 **/
template <class T>
class h_table{
 private:
  int currentIndex;            /** Number of nodes, where *current resides. **/
  table_node<T> *first, *current; /** Nodes allow h_table to be accessed as a list. **/
  table_node<T> **buckets;        /** The table, itself. **/
  unsigned int table_length;      /** Number of buckets. **/


  /** Private Methods: see implementations for details. **/

  table_node<T> *new_node();
  unsigned int hash_str(char *str);
  void rehash();

 public:
  unsigned int size;

  /** Public Methods: see implementations for details. **/

  h_table();
  ~h_table();
  bool contains(T* data);
  int find(T* data);
  T* at(int i);
  void add(T* data);

  /*
  bool remove(T *data);
  */

  void set(int i, T* data);
  T* get_named(char *str);
  void print_stats();
};






#endif
