%{

/**  File:        ug_read.l --> lex.ug_read.c
 **  Created by:  wml
 **  Modified by: wml
 **  Invocation:  flex -Pug_read ug_read.l
 **  Description: Lexical analyzer for files in unigraphics format.
 **/

#include <iostream>
#include <string.h>
#include "ug_read.tab.h"

unsigned int line;                   /** Line number. **/
extern void ug_readerror(char *str); /** yyerror(*str). **/

%}

%option noyywrap

Layer		c_rgb
Color		color
Vertex		v
Wire		w
Face		f
Instance	i
Array		"a "([^\n]+)
Definition	def
End		end
Include		include
Transform	"-"[rst][xyzav]([^\)]+)
Number		(-?)([0-9]+)|(-?)([0-9]*)[.]([0-9]+)
Identifier	[a-zA-Z._?'#@\&\*:$0-9/]([a-zA-Z._?'#@\&\*:0-9$\-/]*)
Semicolon	[;]
OC		"{"
CC		"}"
OS		"["
CS		"]"
OP		[(]
CP		[)]
Newline		[\n\r]+
Whitespace	[ \t]+

%%

{Layer}		{ return TOK_LAYER; }
{Color}		{ return TOK_COLOR; }
{Definition}	{ return TOK_DEFINITION; }
{Include}	{ return TOK_INCLUDE; }
{End}		{ return TOK_END; }
{Vertex}	{ return TOK_VERT; }
{Wire}		{ return TOK_WIRE; }
{Face}		{ return TOK_FACE; }
{Instance}	{ return TOK_INSTANCE; }
{Transform}	{
		  ug_readlval.str = strdup(ug_readtext);
		  return TRANSFORM;
		}
{Array}		{
		  ug_readlval.str = strdup(ug_readtext);
		  return ARRAY;
		}
{OC}		{ return TOK_OC; }
{CC}		{ return TOK_CC; }
{OS}		{ return TOK_OS; }
{CS}		{ return TOK_CS; }
{OP}		{ return TOK_OP; }
{CP}		{ return TOK_CP; }
{Semicolon}	{ return TOK_EOL; }
{Newline}	{ ++line; }
{Whitespace}	/** Do Nothing. **/
{Number}	{
		  //ug_readlval.num = (double) atold(ug_readtext);
                  ug_readlval.num = atof(ug_readtext);
		  return DOUBLE;
		}
{Identifier}	{
		  ug_readlval.str = strdup(ug_readtext);
		  return IDENT;
		}

%%

