#ifndef _DXF_READER_H
#define _DXF_READER_H

#include <CADstuff.h>
#include <dl_creationinterface.h>
#include <map>
#include "ug_builder.h"

class DXF_Reader : public DL_CreationInterface {

public:
  DXF_Reader(UG_Builder *ug_builder);
  virtual ~DXF_Reader();

  
  virtual void addLayer(const DL_LayerData& data);
  virtual void addBlock(const DL_BlockData& data);
  virtual void endBlock();
  virtual void addPoint(const DL_PointData& data);
  virtual void addLine(const DL_LineData& data);
  virtual void addArc(const DL_ArcData& data);
  virtual void addCircle(const DL_CircleData& data);
  virtual void addEllipse(const DL_EllipseData& data);
  virtual void addPolyline(const DL_PolylineData& data);
  virtual void addVertex(const DL_VertexData& data);
  virtual void addInsert(const DL_InsertData& data);
  virtual void addMText(const DL_MTextData& data);
  virtual void addMTextChunk(const char* text);
  virtual void addText(const DL_TextData& data);
  virtual void addDimAlign(const DL_DimensionData& data,
                           const DL_DimAlignedData& edata);
  virtual void addDimLinear(const DL_DimensionData& data,
                            const DL_DimLinearData& edata);
  virtual void addDimRadial(const DL_DimensionData& data,
                            const DL_DimRadialData& edata);
  virtual void addDimDiametric(const DL_DimensionData& data,
                               const DL_DimDiametricData& edata);
  virtual void addDimAngular(const DL_DimensionData& data,
                             const DL_DimAngularData& edata);
  virtual void addDimAngular3P(const DL_DimensionData& data,
                               const DL_DimAngular3PData& edata);
  virtual void addLeader(const DL_LeaderData& data);
  virtual void addLeaderVertex(const DL_LeaderVertexData& data);
  virtual void addHatch(const DL_HatchData& data);
  virtual void addHatchLoop(const DL_HatchLoopData& data);
  virtual void addHatchEdge(const DL_HatchEdgeData& data);

  // not sure when this happens
  virtual void endEntity();
  
  virtual void endSequence();

  virtual void setVariableVector(const char* key,
                                 double v1, double v2, double v3, int code);
  virtual void setVariableString(const char* key,
                                 const char* value, int code);
  virtual void setVariableInt(const char* key, int value, int code);
  virtual void setVariableDouble(const char* key, double value, int code);
 

  /*
  void printAttributes();
  */

 protected:
  /*
  class r {
    ug_region *region;
    ug_layer *layer;
  }
  */

 protected:
  const string getCurrentLayer() const;

 protected:
  UG_Builder *ugb;
  unsigned int polyline_vert_count;

  // indicates that we're currently reading vertices (for a polyline) only
  //   should have similar checks for BLOCK...
  bool reading_polyline;
  
  // we ignore a POLYLINE if it's not on a supported layer
  bool ignoring_polyline;

  map < string, pair<ug_region*, ug_layer*> >regions;
};


#endif
