#ifndef _UG_BUILDER_H
#define _UG_BUILDER_H

// the UG library
#include <CADstuff.h>

class UG_Builder {

 public:
  UG_Builder();
  ~UG_Builder();

  ug_region *beginDefinition(const string &name);
  void endDefinition();
  ug_vert *addAnonymousVertex(double x, double y, double z);
  ug_wire *addAnonymousWire(const string &layer_name);
  bool beginAnonymousWire(const string layer_name);
  void addAnonymousVertexToCurrentWire(double x, double y, double z);
  void endAnonymousWire();
  void addLayer(const string &layer_name);
  void addInstance(const string &region_name,
                   double ox, double oy, double oz,
                   double sx, double sy, double sz,
                   double rz);
  void addAnonymousArc(double cx, double cy, double cz,
                       double r, double a1, double a2,
                       ug_layer *layer);
  void addAnonymousCircle(double x, double y, double z, double r,
                          ug_layer *layer);
  void addAnonymousEllipse(double cx, double cy, double cz,
                           double mx, double my, double mz,
                           double ratio, double a1, double a2,
                           ug_layer *layer);
  void addTextVertex(string text, string layer_name,
                     double ix, double iy, double iz);

  void readLayerFile(char *filename);
  void writeUGFile(const string &filename);

  ug_layer *getLayer(string layer_name);

 protected:
  static int readAlphanumToken(char *b, int &start, int len);
  
  ug_CADSpace ug; // the UG structure
  ug_region *current_region; // the region we're currently building
  int current_color_index; // used for c_rgb in extended UG files
  h_table <ug_layer>valid_layers;
  
  ug_wire *current_poly_wire; // the polyline-based wire (or otherwise)
                              //   currently being built

  // this is the layer for arcs/ellipses/circles
  ug_layer *arc_layer;

  // this is the layer for space names
  //ug_layer *name_layer;
};


#endif
