#include <set>
#include <map>
#include <CADstuff.h>
#include <support.h>
#include <assert.h>

using namespace std;


// used for equal()
#define EPSILON (1E-7)


void print_usage(char *cmdname);

extern void ug_write(ug_CADSpace &uc, char *filename);
extern void ug_parse(ug_CADSpace &uc, char *filename);


inline bool equal(double a, double b)
{
  double d;

  d = a - b;
  return (d >= 0.0 && d < EPSILON) || (d <= 0.0 && d > -EPSILON);
}

class ug_vert_compare {
public:
  bool operator()(ug_vert *a, ug_vert *b) const
  {
    if (!equal(a->x, b->x)) return a->x < b->x;
    if (!equal(a->y, b->y)) return a->y < b->y;
    if (!equal(a->z, b->z)) return a->z < b->z;
    else return false;
  }
};

/*
  checks the name of v -- if it is of the form
    n_<int>_<string>, then it is a text vertex, and
    the text string is returned -- otherwise NULL
    is returned
*/
char *is_text_vert(ug_vert *v)
{
  char *s;
  int i;

  s = v->get_name();
  if (s[0] != 'n' || s[1] != '_') return NULL;
  for (i=2; s[i] != '_' && s[i] != '\0'; i++) {
    if (s[i] < '0' || s[i] > '9') return NULL;
  }
  if (s[i]=='\0') return NULL;

  return s+i+1;
}



int main(int argc, char **argv)
{
  ug_CADSpace ug, ug_out;
  set <ug_vert*, ug_vert_compare>oldVerts;
  map <char*, ug_vert*>undupVerts;
  //set<ug_vert*>::iterator it, jt;
  pair < set<ug_vert*>::iterator, bool > rt;
  ug_vert *vIn, *vNew, *vPrev, *vOut;
  ug_layer *layer;
  ug_wire *wIn, *wOut;
  
  // ug_vert_compare eq;
  char *s;
  int i, j;

  if (argc < 3) {
    print_usage(argv[0]);
    exit(1);
  }

  ug_parse(ug, argv[1]);






  /*
    for now dejunk will:

    - flatten the UG structure
    - remove multiple vertices with the same
      coordinates

    - note: we need to handle text vertices specially!!
  */

  // not 100% sure this works yet
  ug.flatten();



  for (i=0; i < ug.numLayers(); i++) {
    layer = ug.getLayerByNumber(i);
    /*
    layer = ug_out.addNewLayer(layer->get_name(),
                               layer->r, layer->g, layer->b, layer->c);
    */
    layer = ug_out.addExistingLayer(layer);
    assert(layer != NULL);
  }


  for (i=0; i < ug.numVerts(); i++) {

    vIn = ug.getVert(i);

    s = is_text_vert(vIn);
    if (s != NULL) {
      ug_out.addNewNamedVert(s, vIn->x, vIn->y, vIn->z, NULL);
      continue;
    }

    rt = oldVerts.insert(vIn);

    if (!rt.second) {
      /* the insertion failed, so let us take all the wires for v and
         put them on the already-inserted vert
      */
      vPrev = *rt.first;
      undupVerts[vIn->get_name()] = undupVerts[vPrev->get_name()];
    } else {
      // new vertex -- put it in the map

      // adding stuff to the region given by NULL is fine,
      //   since in this flattened structure there are no real regions
      vNew = ug_out.addNewVert(vIn->x, vIn->y, vIn->z, NULL);
      undupVerts[vIn->get_name()] = vNew;
    }
  }

  for (i=0; i < ug.numWires(); i++) {
    wIn = ug.getWire(i);
    wOut = ug_out.addNewWire(wIn->myLayer, NULL);
    assert(wOut != NULL);
    
    for (j=0; j < wIn->numVerts(); j++) {
      vIn = wIn->getVert(j);
      vOut = undupVerts[vIn->get_name()];
      wOut->pushVert(vOut);
    }
  }

  // still need to handle text verts ...

  ug_write(ug_out, argv[2]);

  return 0;
}




void print_usage(char *cmdname)
{
  fprintf(stdout, "usage:\n");
  fprintf(stdout, "%s <UG input file> <UG output file>\n",
          cmdname);
}
